/*
Szymon Rusinkiewicz
Princeton University

mesh_view.cc
Simple viewer
*/

#include <stdio.h>
#include <stdlib.h>
#include "TriMesh.h"
#include "XForm.h"
#include "GLCamera.h"
#include "gl2ps.h"

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <string>
#include <iostream>
#include <sstream>

using std::string;
using std::ofstream;
using std::ostringstream;
using std::endl;
using std::cout;

// Globals
vector<TriMesh *> meshes;
vector<xform> xforms;
vector<bool> visible;
vector<char *> xffilenames;

TriMesh::BSphere global_bsph;
xform global_xf;
GLCamera camera;
int current_mesh = -1;

bool draw_edges = true;
bool draw_2side = true;
bool draw_shiny = true;
bool draw_lit = true;
bool draw_curv = false;
bool draw_falsecolor = false;
bool draw_index = false;
bool non_interactive_mode = false;
int non_interactive_ss = 0;

string lightsfilename;
string sshotfilename;

vector<float> clen;

int tile_count = 1;

// Make some mesh current
void set_current(int i)
{
	if (i >= 0 && i < meshes.size() && visible[i])
		current_mesh = i;
	else
		current_mesh = -1;
	camera.stopspin();
}


// Change visiblility of a mesh
void toggle_vis(int i)
{
	if (i >= 0 && i < meshes.size())
		visible[i] = !visible[i];
	if (current_mesh == i && !visible[i])
		set_current(-1);
}


// Signal a redraw
void need_redraw()
{
	glutPostRedisplay();
}

int current_tile = tile_count * tile_count;

GLfloat bg_color[4] = {0.08,0.08,0.08,0};
GLfloat edge_mat[4] = {0,0,1,1};
GLfloat mat_diffuse[4] = { 1, 1, 1, 1 };
GLfloat mat_specular[4] = { 0.18, 0.18, 0.18, 0.18 };
GLfloat point_size = 4.0;
GLfloat poly_offset = 3.0;
GLfloat line_width = 1.0;

GLfloat light_ambient[8][4] = 
	    {
		{0,0,0,0},
	    };
GLfloat light_diffuse[8][4] = 
	    {
		{0.85,0.85,0.8,0.85},
		{ -0.01, -0.01, -0.03, -0.03 },
		{ 0,0,0,0},
	    };
GLfloat light_specular[8][4] = 
	    {
		{0.85,0.85,0.85,0.85},
		{0,0,0,0},
	    };

int lights[8] = { 1,1,0,0,0,0,0,0 };

void switch_light(int i)
{
    lights[i] = !lights[i];
    if (lights[i]) {
	glEnable(GL_LIGHT0 + i);
	printf("Enabled light %d: %f %f %f %f\n", i, 
	       light_diffuse[i][0], light_diffuse[i][1], light_diffuse[i][2], light_diffuse[i][3]);
    } else {
	glDisable(GL_LIGHT0 + i);
	printf("Disabled light %d\n", i);
    }
}

void change_tiling()
{
    current_tile++;
    if (current_tile > tile_count * tile_count)
	current_tile = 0;
    
    if (current_tile == tile_count * tile_count) {
	camera.setTile(-1, 1, -1, 1);
    } else {
	int x = current_tile % tile_count,
	    y = current_tile / tile_count;
	float delta = 2 / ((float) tile_count);
	float 
	    l = (x * delta)-1.0f, r = ((x+1.0f) * delta)-1.0f,
	    d = (y * delta)-1.0f, u = ((y+1.0f) * delta)-1.0f;
	camera.setTile(l, r, d, u);
    }
}
bool pcl_pointsize(int &argc, char **&argv)
{
    string c = argv[0];
    if (string("-pointsize") == c) {
	point_size = atof(argv[1]);
	argc-=1;
	argv+=1;
	return true;
    } else
	return false;
}

bool pcl_light(int &argc, char **&argv)
{
    string c = argv[0];
    if (string("-light") == c) {
	string lk = argv[1];
	int li = atoi(argv[2]);
	float r = atof(argv[3]),
	    g = atof(argv[4]),
	    b = atof(argv[5]),
	    a = atof(argv[6]);
	if (lk == "diffuse") {
	    light_diffuse[li][0] = r;
	    light_diffuse[li][1] = g;
	    light_diffuse[li][2] = b;
	    light_diffuse[li][3] = a;
	} else if (lk == "ambient") {
	    light_ambient[li][0] = r;
	    light_ambient[li][1] = g;
	    light_ambient[li][2] = b;
	    light_ambient[li][3] = a;
	} else if (lk == "specular") {
	    light_specular[li][0] = r;
	    light_specular[li][1] = g;
	    light_specular[li][2] = b;
	    light_specular[li][3] = a;
	}
	argc -= 6;
	argv += 6;
	return true;
    }
    return false;
}

bool pcl_linewidth(int &argc, char **&argv)
{
    string c = argv[0];
    if (string("-linewidth") == c) {
	line_width = atof(argv[1]);
	argc-=1;
	argv+=1;
	return true;
    } else
	return false;
}

bool pcl_bgcolor(int &argc, char **&argv)
{
    string c = argv[0];
    if (string("-bgcolor") == c) {
	bg_color[0] = atof(argv[1]);
	bg_color[1] = atof(argv[2]);
	bg_color[2] = atof(argv[3]);
	bg_color[3] = atof(argv[4]);
	argc-=4;
	argv+=4;
	return true;
    } else
	return false;
}

bool pcl_edgecolor(int &argc, char **&argv)
{
    string c = argv[0];
    if (string("-edgecolor") == c) {
	edge_mat[0] = atof(argv[1]);
	edge_mat[1] = atof(argv[2]);
	edge_mat[2] = atof(argv[3]);
	edge_mat[3] = atof(argv[4]);
	argc-=4;
	argv+=4;
	return true;
    } else
	return false;
}

bool pcl_diffuse(int &argc, char **&argv)
{
    string c = argv[0];
    if (string("-diffusecolor") == c) {
	mat_diffuse[0] = atof(argv[1]);
	mat_diffuse[1] = atof(argv[2]);
	mat_diffuse[2] = atof(argv[3]);
	mat_diffuse[3] = atof(argv[4]);
	argc-=4;
	argv+=4;
	return true;
    } else
	return false;
}

bool pcl_polyoffset(int &argc, char **&argv)
{
    string c = argv[0];
    if (string("-polyoffset") == c) {
	poly_offset = atof(argv[1]);
	argc-=1;
	argv+=1;
	return true;
    } else
	return false;
}

bool pcl_specular(int &argc, char **&argv)
{
    string c = argv[0];
    if (string("-specularcolor") == c) {
	mat_specular[0] = atof(argv[1]);
	mat_specular[1] = atof(argv[2]);
	mat_specular[2] = atof(argv[3]);
	mat_specular[3] = atof(argv[4]);
	argc-=4;
	argv+=4;
	return true;
    } else
	return false;
}

bool pcl_noninteractive(int &argc, char **&argv)
{
	string c = argv[0];
	if (string("-noninteractive") == c) {
		non_interactive_mode = true;
		non_interactive_ss = atoi(argv[1]);
		argc--;
		argv++;
		return true;
	} else
		return false;
}

void process_command_line(int argc, char **argv)
{
    while (argc > 0) {
	pcl_bgcolor(argc, argv) ||
	    pcl_edgecolor(argc, argv) ||
	    pcl_diffuse(argc, argv) ||
	    pcl_specular(argc, argv) ||
	    pcl_linewidth(argc, argv) ||
	    pcl_pointsize(argc, argv) ||
	    pcl_light(argc, argv) ||
	    pcl_polyoffset(argc, argv) ||
            pcl_noninteractive(argc, argv);
	--argc; ++argv;
    }
}

void cls()
{
	glDisable(GL_DITHER);
	glDisable(GL_BLEND);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_NORMALIZE);
	glDisable(GL_LIGHTING);
	glDisable(GL_NORMALIZE);
	glDisable(GL_COLOR_MATERIAL);
	glClearColor(bg_color[0], bg_color[1], bg_color[2], bg_color[3]);
	glClearDepth(1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

}

// Set up lights and materials
void setup_lighting(int mesh)
{
	if (!draw_lit || meshes[mesh]->normals.empty()) {
		glDisable(GL_LIGHTING);
		glColor4fv(mat_diffuse);
		return;
	}

	// setup materials
	if (!draw_shiny) {
		mat_specular[0] = mat_specular[1] =
		mat_specular[2] = mat_specular[3] = 0.0f;
	} else {
		mat_specular[0] = mat_specular[1] =
		mat_specular[2] = mat_specular[3] = 0.18f;
	}
	GLfloat mat_shininess[] = { 64 };
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, mat_diffuse);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat_specular);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, mat_shininess);
	GLfloat global_ambient[] = { 0.02, 0.02, 0.05, 0.05 };

	for (int i=0; i<8; ++i) {
	    glLightfv(GL_LIGHT0 + i, GL_AMBIENT, light_ambient[i]);
	    glLightfv(GL_LIGHT0 + i, GL_DIFFUSE, light_diffuse[i]);
	    glLightfv(GL_LIGHT0 + i, GL_SPECULAR, light_specular[i]);
	}
				       
// 	GLfloat light0_ambient[] = { 0, 0, 0, 0 };
// 	GLfloat light0_diffuse[] = { 0.85, 0.85, 0.8, 0.85 };
// 	GLfloat light1_diffuse[] = { -0.01, -0.01, -0.03, -0.03 };
// 	GLfloat light0_specular[] = { 0.85, 0.85, 0.85, 0.85 };
// 	glLightfv(GL_LIGHT0, GL_AMBIENT, light0_ambient);
// 	glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
// 	glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
// 	glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);

	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_FALSE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, draw_2side);
	if (!meshes[mesh]->colors.empty()) {
		glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
		glEnable(GL_COLOR_MATERIAL);
	}
	glEnable(GL_LIGHTING);
	for (int i=0; i<8; ++i)
	    if (lights[i])
		glEnable(GL_LIGHT0 + i);
	glEnable(GL_NORMALIZE);
}


// Draw triangle strips.  They are stored as length followed by values.
void draw_tstrips(const TriMesh *themesh)
{
	const int *t = &themesh->tstrips[0];
	const int *end = t + themesh->tstrips.size();
	while (likely(t < end)) {
		int striplen = *t++;
		glDrawElements(GL_TRIANGLE_STRIP, striplen, GL_UNSIGNED_INT, t);
		t += striplen;
	}
}

// Draw the mesh
void draw_mesh(int i)
{
	TriMesh *themesh = meshes[i];
	int mesh_number = i;

	glPushMatrix();
	glMultMatrixd(xforms[i]);

	glDepthFunc(GL_LESS);
	glEnable(GL_DEPTH_TEST);

	if (draw_2side) {
		glDisable(GL_CULL_FACE);
	} else {
		glCullFace(GL_BACK);
		glEnable(GL_CULL_FACE);
	}

	// Vertices
	glEnableClientState(GL_VERTEX_ARRAY);
	glVertexPointer(3, GL_FLOAT,
			sizeof(themesh->vertices[0]),
			&themesh->vertices[0][0]);
	// Normals
	if (!themesh->normals.empty() && !draw_index) {
		glEnableClientState(GL_NORMAL_ARRAY);
		glNormalPointer(GL_FLOAT,
				sizeof(themesh->normals[0]),
				&themesh->normals[0][0]);
	} else {
		glDisableClientState(GL_NORMAL_ARRAY);
	}

	// Colors
	if (!themesh->colors.empty() && !draw_falsecolor && !draw_index) {
		glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(3, GL_UNSIGNED_BYTE,
			       sizeof(themesh->colors[0]),
			       &themesh->colors[0][0]);
		glEnable(GL_COLOR_MATERIAL);
	} else {
		glDisableClientState(GL_COLOR_ARRAY);
	}

	// Main drawing pass
	if (themesh->tstrips.empty()) {
		// No triangles - draw as points
		glPointSize(point_size * tile_count);
		glDrawArrays(GL_POINTS, 0, themesh->vertices.size());
		glPopMatrix();
		return;
	}

	if (draw_edges || draw_curv) {
		glPolygonOffset(poly_offset, poly_offset);
		glEnable(GL_POLYGON_OFFSET_FILL);
	}

	draw_tstrips(themesh);
	glDisable(GL_POLYGON_OFFSET_FILL);

	// Edge drawing pass
	if (draw_edges) {
		glLineWidth(line_width);
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		glDisableClientState(GL_COLOR_ARRAY);
		glDisable(GL_COLOR_MATERIAL);
		GLfloat global_ambient[] = { 0.2, 0.2, 0.2, 1.0 };
		GLfloat light0_diffuse[] = { 0.8, 0.8, 0.8, 0.0 };
		GLfloat light1_diffuse[] = { -0.2, -0.2, -0.2, 0.0 };
		GLfloat light0_specular[] = { 0.0f, 0.0f, 0.0f, 0.0f };
		glLightModelfv(GL_LIGHT_MODEL_AMBIENT, global_ambient);
		glLightfv(GL_LIGHT0, GL_DIFFUSE, light0_diffuse);
		glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
		glLightfv(GL_LIGHT0, GL_SPECULAR, light0_specular);
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, edge_mat);
		glColor4fv(edge_mat); // Used iff unlit
		draw_tstrips(themesh);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	}

	if (draw_curv) {
			glDisable(GL_LIGHTING);
			themesh->need_curvatures();
			int nv = themesh->vertices.size();
			glColor3f(1,0,0);
			glBegin(GL_LINES);
			int i;
			for (i = 0; i < nv; i++) {
				glVertex3fv(themesh->vertices[i]);
				glVertex3fv(themesh->vertices[i] +
					    clen[mesh_number] * themesh->normals[i]);
			}
			glColor3f(0.7,0.7,0);
			for (i = 0; i < nv; i++) {
				glVertex3fv(themesh->vertices[i] -
					    clen[mesh_number] * themesh->pdir1[i]);
				glVertex3fv(themesh->vertices[i] +
					    clen[mesh_number] * themesh->pdir1[i]);
			}
			glColor3f(0,1,0);
			for (i = 0; i < nv; i++) {
				glVertex3fv(themesh->vertices[i] -
					    clen[mesh_number] * themesh->pdir2[i]);
				glVertex3fv(themesh->vertices[i] +
					    clen[mesh_number] * themesh->pdir2[i]);
			}
			glEnd();
	}

	glPopMatrix();
}

// Draw the scene
void redraw()
{
	timestamp t = now();
	camera.setupGL(global_xf * global_bsph.center, global_bsph.r);
	glPushMatrix();
	glMultMatrixd(global_xf);
	cls();
	for (int i = 0; i < meshes.size(); i++) {
		if (!visible[i])
			continue;
		setup_lighting(i);
		draw_mesh(i);
	}
	glPopMatrix();
	glutSwapBuffers();
	printf("\r                        \r%.1f msec.", 1000.0f * (now() - t));
	fflush(stdout);
}

void write_print_quality_screenshot(int tc)
{
    tc = 1 << tc;
    int otc = tile_count;
    int oct = current_tile;
    tile_count = tc;
    current_tile = tc * tc;

    glLineWidth(line_width * (float)tile_count);

    GLint V[4];
    glGetIntegerv(GL_VIEWPORT, V);
    int width = V[2], height = V[3];

    unsigned char *fb = new unsigned char[width * height * 3];
    printf("%d %d\n", width, height);

    printf("%d %d\n", width, height);
    for (int row=0; row<tile_count; ++row)
	for (int col=0; col<tile_count; ++col) {
	    FILE *f;
	    char fname[1024];
	    sprintf(fname, "sshot.ppm");
	    f = fopen(fname, "w");
	    fprintf(f, "P6\n%d %d\n255\n", width, height);
	    printf("                   \r%d %d", row, col);
	    change_tiling();
	    redraw();
	    glReadBuffer(GL_FRONT);
	    glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, fb);
	    /*	    for (int x=0; x<width; ++x)
		for (int y=0; y<height; ++y) {
		    int fcolumn = x + col * width,
			frow = height - (y + row * height)-1;
		    vs[3*(frow * width + fcolumn)]   = fb[3*(y*width+x)];
		    vs[3*(frow * width + fcolumn)+1] = fb[3*(y*width+x)+1];
		    vs[3*(frow * width + fcolumn)+2] = fb[3*(y*width+x)+2];
		    }*/
	    fwrite(fb, 1, width*height*3, f);
	    fclose(f);
	    char commandline[1024];
	    sprintf(commandline, "convert sshot.ppm -flip -scale %dx%d sshot_%d_%d.png", 
		    width/tile_count,
		    height/tile_count,
		    row, col);
	    system(commandline);
	    sprintf(commandline, "rm sshot.ppm");
	    system(commandline);
	}

    //    for (int i=0; i<width*height*tile_count*tile_count*3; ++i)
    //	fprintf(f, "%d\n", vs[i]);

    delete fb;

    ostringstream s;
    s << "montage";
    for (int row = tile_count-1; row >= 0; --row)
	for (int col = 0; col < tile_count; ++col)
	    s << " " << "sshot_" << row << "_" << col << ".png";
    s << " -geometry " << width/tile_count << "x" << height/tile_count << " -tile " << tile_count <<"x"<< tile_count;
    s << " " << sshotfilename;
    std::cout << "\"" << s.str() << "\"" << endl;
    system(s.str().c_str());
    for (int row = 0; row < tile_count; ++row)
	for (int col = 0; col < tile_count; ++col) {
	    char commandline[1024];
	    sprintf(commandline, "rm sshot_%d_%d.png", row, col);
	    system(commandline);
	    }
    ostringstream s2;
    if (!non_interactive_mode) {
	    s2 << "feh " << sshotfilename << "&";
	    system(s2.str().c_str());
    }
    glLineWidth(line_width);

    change_tiling();
    tile_count = otc;
    current_tile = current_tile;
}

void usage(const char *myname)
{
	fprintf(stderr, "Usage: %s infile [opts]\n", myname);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "-bgcolor r g b a: background color\n");
	fprintf(stderr, "-edgecolor r g b a: edge color\n");
	fprintf(stderr, "-diffusecolor r g b a: diffuse material\n");
	fprintf(stderr, "-specularcolor r g b a: specular material\n");
	fprintf(stderr, "-pointsize s: size of point in pixels (fractional works for ss)\n");
	fprintf(stderr, "-linewidth s: width of line in pixels (fractional works for ss)\n");
	exit(1);
}

// Update global bounding sphere.
void update_bsph()
{
	point boxmin(1e38, 1e38, 1e38);
	point boxmax(-1e38, -1e38, -1e38);
	bool some_vis = false;
	for (int i = 0; i < meshes.size(); i++) {
		if (!visible[i])	
			continue;
		some_vis = true;
		point c = xforms[i] * meshes[i]->bsphere.center;
		float r = meshes[i]->bsphere.r;
		for (int j = 0; j < 3; j++) {
			boxmin[j] = min(boxmin[j], c[j]-r);
			boxmax[j] = max(boxmax[j], c[j]+r);
		}
	}
	if (!some_vis)
		return;
	point &gc = global_bsph.center;
	float &gr = global_bsph.r;
	gc = 0.5f * (boxmin + boxmax);
	gr = 0.0f;
	for (int i = 0; i < meshes.size(); i++) {
		if (!visible[i])	
			continue;
		point c = xforms[i] * meshes[i]->bsphere.center;
		float r = meshes[i]->bsphere.r;
		gr = max(gr, dist(c, gc) + r);
	}
}


// Set the view...
void resetview()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	camera.stopspin();
	for (int i = 0; i < meshes.size(); i++)
		if (!xforms[i].read(xffilenames[i])) {
			xforms[i] = xform();
		} else {
			//std::cout << "READ" << std::endl;
		}

	// Special case for 1 mesh
	if (meshes.size() == 1 && xforms[0].read(xffilenames[0])) {
		global_xf = xforms[0];
		xforms[0] = xform();
		update_bsph();
	} else {
		update_bsph();
		global_xf = xform::trans(0, 0, -5.0f * global_bsph.r) *
			xform::trans(-global_bsph.center);
	}

// 	cout << global_xf << endl;
// 	for (int i = 0; i < meshes.size(); ++i) {
// 		cout << xforms[i] << endl;
// 	}
}


// Save the current image to a PPM file.
// Uses the next available filename matching filenamepattern
void dump_image()
{
	// Find first non-used filename
	const char filenamepattern[] = "img%d.ppm";
	int imgnum = 0;
	FILE *f;
	while (1) {
		char filename[1024];
		sprintf(filename, filenamepattern, imgnum++);
		f = fopen(filename, "rb");
		if (!f) {
			f = fopen(filename, "wb");
			printf("\n\nSaving image %s... ", filename);
			fflush(stdout);
			break;
		}
		fclose(f);
	}

	// Read pixels
	GLint V[4];
	glGetIntegerv(GL_VIEWPORT, V);
	GLint width = V[2], height = V[3];
	char *buf = new char[width*height*3];
	glPixelStorei(GL_PACK_ALIGNMENT, 1);
	glReadPixels(V[0], V[1], width, height, GL_RGB, GL_UNSIGNED_BYTE, buf);

	// Flip top-to-bottom
	for (int i = 0; i < height/2; i++) {
		char *row1 = buf + 3 * width * i;
		char *row2 = buf + 3 * width * (height - 1 - i);
		for (int j = 0; j < 3 * width; j++)
			swap(row1[j], row2[j]);
	}

	// Write out file
	fprintf(f, "P6\n#\n%d %d\n255\n", width, height);
	fwrite(buf, width*height*3, 1, f);
	fclose(f);
	delete [] buf;

	printf("Done.\n\n");
}


// Handle mouse button and motion events
static unsigned buttonstate = 0;

void doubleclick(int button, int x, int y)
{
	if (button == 0 || buttonstate == (1 << 0)) {
		// Double left click - select a mesh
		camera.setupGL(global_xf * global_bsph.center, global_bsph.r);
		glDisable(GL_BLEND);
		glDisable(GL_LIGHTING);
		glClearColor(1,1,1,1);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glEnable(GL_DEPTH_TEST);
		draw_index = true;
		glPushMatrix();
		glMultMatrixd(global_xf);
		for (int i = 0; i < meshes.size(); i++) {
			if (!visible[i])
				continue;
			glColor3ub((i >> 16) & 0xff,
				   (i >> 8)  & 0xff,
				    i        & 0xff);
			draw_mesh(i);
		}
		glPopMatrix();
		draw_index = false;
		GLint V[4];
		glGetIntegerv(GL_VIEWPORT, V);
		y = int(V[1] + V[3]) - 1 - y;
		unsigned char pix[3];
		glReadPixels(x, y, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, pix);
		int n = (pix[0] << 16) + (pix[1] << 8) + pix[2];
		set_current(n);
	}
}

void mousemotionfunc(int x, int y)
{
	static const Mouse::button physical_to_logical_map[] = {
		Mouse::NONE, Mouse::ROTATE, Mouse::MOVEXY, Mouse::MOVEZ,
		Mouse::MOVEZ, Mouse::MOVEXY, Mouse::MOVEXY, Mouse::MOVEXY,
	};

	Mouse::button b = Mouse::NONE;
	if (buttonstate & (1 << 3))
		b = Mouse::WHEELUP;
	else if (buttonstate & (1 << 4))
		b = Mouse::WHEELDOWN;
	else if (buttonstate & (1 << 30))
		b = Mouse::LIGHT;
	else
		b = physical_to_logical_map[buttonstate & 7];

	if (current_mesh < 0) {
		camera.mouse(x, y, b,
			     global_xf * global_bsph.center, global_bsph.r,
			     global_xf);
	} else {
		xform tmp_xf = global_xf * xforms[current_mesh];
		camera.mouse(x, y, b,
			     tmp_xf * meshes[current_mesh]->bsphere.center,
			     meshes[current_mesh]->bsphere.r,
			     tmp_xf);
		xforms[current_mesh] = inv(global_xf) * tmp_xf;
		update_bsph();
	}
	if (b != Mouse::NONE)
		need_redraw();
}

void mousebuttonfunc(int button, int state, int x, int y)
{
	static timestamp last_click_time;
	static unsigned last_click_buttonstate = 0;
	static float doubleclick_threshold = 0.25f;

	if (glutGetModifiers() & GLUT_ACTIVE_CTRL)
		buttonstate |= (1 << 30);
	else
		buttonstate &= ~(1 << 30);

	if (state == GLUT_DOWN) {
		buttonstate |= (1 << button);
		if (buttonstate == last_click_buttonstate &&
		    now() - last_click_time < doubleclick_threshold)
			doubleclick(button, x, y);
		last_click_time = now();
		last_click_buttonstate = buttonstate;
	} else {
		buttonstate &= ~(1 << button);
	}

	mousemotionfunc(x, y);
}


// Idle callback
void idle()
{
	xform tmp_xf = global_xf;
	if (current_mesh >= 0)
		tmp_xf = global_xf * xforms[current_mesh];

	if (camera.autospin(tmp_xf))
		need_redraw();
	else
		usleep(10000);

	if (current_mesh >= 0) {
		xforms[current_mesh] = inv(global_xf) * tmp_xf;
		update_bsph();
	} else {
		global_xf = tmp_xf;
	}
}


// Keyboard
void keyboardfunc(unsigned char key, int x, int y)
{
    enum KSMState {
	Start, PrintSShot, Light
    };
    static KSMState state = Start;

    key = tolower(key);

    switch (state) {
    case Start: {
	switch (key) {
		case ' ':
			if (current_mesh < 0)
				resetview();
			else
				set_current(-1);
			break;
		case '@': // Shift-2
			draw_2side = !draw_2side; break;
		case 'e':
			draw_edges = !draw_edges; break;
		case 'f':
			draw_falsecolor = !draw_falsecolor; break;
		case 'c':
		    draw_curv = !draw_curv; break;
		case 'l': 
			draw_lit = !draw_lit; break;
		case 12: // Ctrl +l
		    state = Light;
		    break;
		case 's':
		    draw_shiny = !draw_shiny; break;
	case 'x': {
// 		cout << endl << meshes.size() << endl;
// 		cout << "FOO!" << endl;
// 		cout << global_xf << endl;
		if (meshes.size() == 1) {
			global_xf.write(xffilenames[0]);
		} else {
			for (int i = 0; i < meshes.size(); ++i) {
				xforms[i].write(xffilenames[i]);
			}
		}
	} break;
		case 'I':
			dump_image(); break;
		case '\033': // Esc
		case '\021': // Ctrl Q
		case 'Q':
		case 'q':
	    	exit(0);
		case '+':
	    	printf("+\n");
		    point_size *= 1.25; 
	    	line_width *= 1.25; break;
		case '-':
		    printf("-\n");
	    	point_size /= 1.25; 
	    	line_width /= 1.25; break;
		case 'p':
		    state = PrintSShot; break;
		}
 		default:
			if (key >= '1' && key <= '9') {
				int m = key - '1';
				toggle_vis(m);
			}
   } break;
    case PrintSShot: {
	state = Start;
	switch (key) {
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    write_print_quality_screenshot(key - '0'); break;
	case '0':
	    write_print_quality_screenshot(10); break;
		break;
	} 
    } break;
    case Light: {
	state = Start;
	switch (key) {
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	    switch_light(key - '0');
	    break;
	}
    } break;
    }

    need_redraw();
}

void killthis()
{
	write_print_quality_screenshot(1);
	exit(0);
}

int main(int argc, char **argv)
{
	glutInitWindowSize(1024,1024);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutInit(&argc, argv);

	if (argc < 2)
		usage(argv[0]);

	lightsfilename = argv[1];
	lightsfilename += ".lights";

	sshotfilename = argv[1];
	sshotfilename += ".png";

	--argc; ++argv;
	while (argc) {
		const char *filename = argv[0];
		if (argv[0][0] == '-')
			break;
		TriMesh *themesh = NULL;
		try {
			themesh = TriMesh::read(filename);
		} catch (...) {
			printf("asdkljahsjdkh\n");
		}
// 		TriMesh *themesh = TriMesh::read(filename);
		if (!themesh)
			break;
		if (!themesh->faces.empty() || !themesh->tstrips.empty())
			themesh->need_normals();
		themesh->need_tstrips();
		themesh->need_bsphere();
		meshes.push_back(themesh);

		char *xffilename = new char[strlen(filename) + 4];
		strcpy(xffilename, filename);
		char *dot = strrchr(xffilename, '.');
		if (!dot)
			dot = strrchr(xffilename, '\0');
		strcpy(dot, ".xf");
		xffilenames.push_back(xffilename);

		xforms.push_back(xform());
		visible.push_back(true);
		--argc; ++argv;
	}

// 	printf("xf: %s\n", xffilename.c_str());
// 	printf("lights: %s\n", lightsfilename.c_str());
// 	printf("sshot: %s\n", sshotfilename.c_str());
	process_command_line(argc, argv);

	for (int i = 0; i < meshes.size(); ++i) {
		clen.push_back(0.5f * meshes[0]->feature_size());
	}

	glutCreateWindow("mesh_view");
	glutMouseFunc(mousebuttonfunc);
	glutMotionFunc(mousemotionfunc);
	glutDisplayFunc(redraw);
	glutKeyboardFunc(keyboardfunc);
	glutIdleFunc(idle);

	resetview();
	glutPostRedisplay();
	if (non_interactive_mode) {
		glutIdleFunc(killthis);
	}
	glutMainLoop();
}
